/*

  P-Code interpreter (to run the apple pascal system)
  Copyright (C) 2000 Mario Klebsch

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>

#include "psystem.h"
#include "Memory.h"
#include "Diskio.h"

word Syscom;

int List=0;
int Text=0;

void warning(char *Msg, ...)
{
  va_list ap;
  char	Buffer[512];
  va_start(ap, Msg);
  vsnprintf(Buffer, sizeof(Buffer), Msg, ap);
  va_end(ap);
  fprintf(stderr,"warning: %s\n", Buffer);
  /*  TraceLevel=0x7fff;*/
}

void XeqError(word err)
{
  fprintf(stderr,"XeqError: %d\n", err);
  exit(-1);
}

void DoList(word Unit)
{
  int i;
  UnitRead(Unit, 0x100, 0, 2048, 2, 0);
  if (MemRd(IORSLT))
    {
      fprintf(stderr,"DoList: Directory read error\n");
      exit(1);
    }

  for (i=0;i<MemRd(WordIndexed(0x100, 8));i++)
    {
      word Entry=WordIndexed(0x100, 13+13*i);
      int len;
      for (len=0; len<MemRdByte(WordIndexed(Entry, 3),0); len++)
	{
	  char ch=MemRdByte(WordIndexed(Entry, 3),1+len);
	  if (isupper(ch))
	    ch=tolower(ch);
	  putchar(ch);
	}
      putchar('\n');
    }
}

word LookupFile(word Unit, const char *Name, word *Offset,
		word *Len, word *Last)
{
  int i;
  UnitRead(Unit, 0x100, 0, 2048, 2, 0);
  if (MemRd(IORSLT))
    return(0);

  for (i=0;i<MemRd(WordIndexed(0x100, 8));i++)
    {
      word Entry=WordIndexed(0x100, 13+13*i);
      int len;
      for (len=0; len<MemRdByte(WordIndexed(Entry, 3),0); len++)
	if (toupper(MemRdByte(WordIndexed(Entry, 3),1+len)) !=
	    toupper(Name[len]))
	  goto next;
      if (Name[len])
	continue;
	
      *Offset=MemRd(WordIndexed(Entry,0));
      *Len=MemRd(WordIndexed(Entry,1));
      *Last=MemRd(WordIndexed(Entry,11));

      return(1);
    next:
    }
  return(0);
}

int main(int argc, char *argv[])
{
  int i;

  Syscom=WordIndexed(0xfffe, -SYSCOM_SIZE);

  while ((i=getopt(argc, argv, "slt"))!=EOF)
    switch(i)
      {
      case 's':
	SwapBytes=1;
	break;
      case 'l':
	List=1;
	break;
      case 't':
	Text=1;
	break;
      }

  if (optind<argc)
    DiskMount(4, argv[optind++], ReadOnly);
  else
    {
      fprintf(stderr,"%s: Needs filename argument\n", argv[0]);
      exit(1);
    }

  if (List)
    {
      DoList(4);
      exit(0);
    }
  if (optind<argc)
    {
      word Offset, End, Last;
      char *name=argv[optind++];
      if (!LookupFile(4, name, &Offset, &End, &Last))
	{
	  fprintf(stderr, "%s: File not found\n", name);
	  exit(1);
	}

      if (Text)
	Offset +=2;

      while (Offset<End)
	{
	  int Size=(End-Offset-1)*512+Last;
	  if (Size>1024)
	    Size=1024;
	  UnitRead(4, 0x100, 0, Size, Offset, 0);
	  for (i=0; i<Size; i++)
	    {
	      char ch=MemRdByte(0x100,i);
	      if (Text)
		{
		  if (ch==0x10)
		    {
		      i++;
		      for (ch=MemRdByte(0x100,i); ch>0x20; ch--)
			write(1," ", 1);
		    }
		  else if (ch==0)
		    break;
		  else
		    {
		      if (ch==0x0d)
			ch='\n';
		      write(1, &ch, 1);
		    }
		}
	      else
		write(1, &ch, 1);
	    }
	  Offset+=2;
	}
    }
  else
    {
      fprintf(stderr,"%s: Needs filename argument\n", argv[0]);
      exit(1);
    }
  exit(0);
}
